#include "lcd.h"
#include "flappy.h"

ISR(TIMER1_COMPA_vect){
    /* counter salveaza numarul de apeluri a acestei intreruperi */
	/* o intrerupere la 4 secunde */
	if (in_game && difficulty == HARD) {
		timer_counter++;
		/* dupa MAX_TIME secunde jucatorul va pierde */
		if (timer_counter == MAX_TIME / 4) {
			timer_counter = 0;	
			time_over = 1;
		}
	}
}

ISR(PCINT3_vect) {
    if (!(PIND & (1 << MOVE_BTN)) && in_game) {
        /* make the bird jump */
		bird.y -= 5 * speed;
		_delay_ms(75);
    }
}

/* generate buzzer sound */
void point_sound() {
	DDRD |= (1 << BUZZER_BTN);
	
	PIND &= ~(1 << BUZZER_BTN);
	_delay_ms(10);
	PIND |= (1 << BUZZER_BTN);
	_delay_ms(40);
	
	PIND &= ~(1 << BUZZER_BTN);
	_delay_ms(25);
	PIND |= (1 << BUZZER_BTN);
	_delay_ms(25);
	
	PIND &= ~(1 << BUZZER_BTN);
	_delay_ms(40);
	PIND |= (1 << BUZZER_BTN);
	
	DDRD &= ~(1 << BUZZER_BTN);
}
/* generate different buzzer sound */
void lose_sound() {
	DDRD |= (1 << BUZZER_BTN);
	
	for(int i = 0; i < 100; i++) {
		PIND &= ~(1 << BUZZER_BTN);
		_delay_ms(1);
		PIND |= (1 << BUZZER_BTN);
		_delay_ms(3);
	}
	
	DDRD &= ~(1 << BUZZER_BTN);
}
/* initializations */
void init_buttons() {
	DDRD |= (1 << MOVE_BTN);
	PORTD |= (1 << MOVE_BTN);
	
	DDRD |= (1 << SELECT_BTN);
	PORTD |= (1 << SELECT_BTN);
	
	sei();
	// Activez PIN CHANGE INTERRUPT ENABLE pentru PCINT24:31 (portul D)
	PCICR |= (1 << PCIE3);
	// Activez intreruperea pentru butonul PD2
	PCMSK3 |= (1 << PCINT26);
}
void init_bird() {
	bird.x = 2;
	bird.y = LCD_HEIGHT / 2 - BIRD_HEIGTH / 2;
}
void init_timer() {
	ICR1 = 65534;
	OCR1A = 62500;
	TIMSK1 = (1 << OCIE1A);
	TCCR1B = (5 << CS10) | (1 << WGM12) | (1 << WGM13);
}

/* print the bird's pixels */
void display_bird() {
	int i, j;
	int ii, jj;
	for (i = bird.x, ii = 0; i < bird.x + BIRD_LENGTH; i++, ii++) {
		for (j = bird.y, jj = 0; j < bird.y + BIRD_HEIGTH; j++, jj++) {
			if (bird_mat[jj][ii])
				lcd_draw_pixel(i, j);
		}
	}
}
/* generate random height for new pipe */
void new_pipe(int index) {
	pipe_x[index] = LCD_WIDTH - PIPE_WIDTH;
	pipe_h[index] = rand() % (LCD_HEIGHT - PIPE_SPACE - 3) + 3;
}
/* print the pipe's pixels */
void display_pipe(int index) {
	int i, j;
	/* top part of the pipe */
	int x = pipe_x[index];
	int h = pipe_h[index];
	for (i = x; i < x + PIPE_WIDTH; i++) {
		for (j = 0; j < h; j++) {
			lcd_draw_pixel(i, j);
		}
	}
	/* bottom part of the pipe */
	for (i = x; i < x + PIPE_WIDTH; i++) {
		for (j = h + PIPE_SPACE; j < LCD_HEIGHT; j++) {
			lcd_draw_pixel(i, j);
		}
	}
}

/* shows the possible game options */
void main_screen() {
	lcd_clear();
	lcd_update();
	
	lcd_goto_xy(8, 0);
	lcd_str("Flappy Bird");
	
	lcd_goto_xy(8, 1);
	lcd_chr(ARROW_LEFT);
	lcd_str(" PM 2019 ");
	lcd_chr(ARROW_RIGHT);
	
	lcd_goto_xy(0, 3);
	if (option == NEW_GAME) {
		lcd_chr(ARROW_RIGHT);
		lcd_str("New Game");
	} else {
		lcd_str(" New Game");
	}
	lcd_goto_xy(0, 4);
	if (option == BOARD) {
		lcd_chr(ARROW_RIGHT);
		lcd_str("Leaderboard");
	} else {
		lcd_str(" Leaderboard");
	}
	
	lcd_goto_xy(0,5);
	lcd_chr(ARROW_LEFT);
	lcd_str("Move");
	lcd_str("  ");
	lcd_str("Select");
	lcd_chr(ARROW_RIGHT);
}

/* Lost game screen */
void end_screen() {
	in_game = 0;
	
	lcd_clear();
	lcd_update();
	
	lcd_goto_xy(12, 0);
	if (time_over) {
		time_over = 0;
		lcd_str("Time Over");
	} else
		lcd_str("Game Over");
	
	char points[10];
	itoa(score, points, 10);
	lcd_goto_xy(12, 1);
	lcd_str("SCORE:");
	lcd_str(points);
	
	int your_place = -1;
	/* update leaderboard if it is the case */
	for (size_t i = 0; i < 5; i++) {
		int x = read_eeprom_word(&eeprom[i]);
		if (score > x) {
			write_eeprom_word(&eeprom[i], score);
			your_place = i + 1;
			/* shift remaining scores */
			for (size_t j = i + 1; j < 5; j++) {
				int aux = read_eeprom_word(&eeprom[j]);
				write_eeprom_word(&eeprom[j], x);
				x = aux;
			}
		}
	}
	
	if (your_place != -1) {
		lcd_goto_xy(0,2);
		lcd_str("Congratulation");
		lcd_goto_xy(12,3);
		lcd_str("You are in");
		lcd_goto_xy(30,4);
		char place[10];
		itoa(your_place, place, 10);
		lcd_str(place);
		if (your_place == 1)
			lcd_str("st");
		else if (your_place == 2)
			lcd_str("nd");
		else if (your_place == 3)
			lcd_str("rd");
		else
			lcd_str("th");
		lcd_goto_xy(28,5);
		lcd_str("Place");
		
	} else {
		lcd_goto_xy(0,2);
		lcd_str("You didn't get");
		lcd_goto_xy(8,3);
		lcd_str("In the top 5");
	}
	_delay_ms(500);
	
	while(1) { 
		if (!(PIND & (1 << SELECT_BTN))) {
			option = 1;
			_delay_ms(250);
			select_mode();
			return;
		}
	}
}

/* leaderboard was selected */
void leaderboard() {
	lcd_clear();
	lcd_update();
	
	lcd_goto_xy(0, 0);
	lcd_chr(ARROW_LEFT);
	lcd_str("Leaderboard");
	lcd_chr(ARROW_RIGHT);
	
	for (size_t i = 1; i < 6; i++) {
		int x = read_eeprom_word(&eeprom[i - 1]);
		char points[10];
		char index[10];
		itoa(x, points, 10);
		itoa(i, index, 10);
		lcd_goto_xy(0, i);
		lcd_str(index);
		lcd_str(". ");
		lcd_str(points);	
		lcd_str(" points");
	}
	
	_delay_ms(500);
	
	while(1) { 
		if (!(PIND & (1 << SELECT_BTN))) {
			option = 1;
			_delay_ms(250);
			select_mode();
			return;
		}
	}
}

/* select the players option */
void select_mode() {
	main_screen();
	int ok = 1;
	while(1) {
		/* check if player selected action */
		if (!(PIND & (1 << SELECT_BTN))) {
			if (option == NEW_GAME) {
				_delay_ms(250);
				select_difficulty();
			} else
				leaderboard();
			return;
		}
		
		/* check if player changes option */
		if (!(PIND & (1 << MOVE_BTN)) && ok) {
			ok = 0;
			option *= -1;
			main_screen();
			_delay_ms(250);
			ok = 1;
		}
		
	}
	
}

/* check if bird hits a pipe */
int check_collision() {
	int x1 = pipe_x[pipe_start];
	int h1 = pipe_h[pipe_start];
	
	/* check if bird touched the left-most pipe */
	if (bird.x + BIRD_LENGTH >= x1 && bird.x <= x1 + PIPE_WIDTH) {
		if (bird.y <= h1 - eps || bird.y + BIRD_HEIGTH >= h1 + PIPE_SPACE + eps) {
			return 1;
		}
	}
	/* check if bird touched ground */
	if (bird.y >= LCD_HEIGHT - BIRD_HEIGTH) {
		lose_sound();
		end_screen();
		return 1;
	}
	/* check if bird touched floor */
	if (bird.y <= 0) {
		lose_sound();
		end_screen();
		return 1;
	}
	return 0;
}

/* game logic */
void game() {
	size_t i;
		
	score = 0;
	turn = -1;
	pipe_cnt = 0;
	pipe_start = 0;
	in_game = 1;
	timer_counter = 0;
	time_over = 0;
	
	lcd_clear();
	lcd_update();
	
	init_bird();
	display_bird();
	
	while(1) {
		turn++;
		
		/* generate new pipe */
		if (turn % PIPE_TURN == 0) {
			new_pipe(pipe_cnt);
			pipe_cnt++;
		}
		
		/* check if bird collides with anything */
		if (check_collision()) {
			lose_sound();
			end_screen();
			return;
		}
		
		/* check if time expired */
		if (time_over) {
			lose_sound();
			end_screen();
			return;
		}
		
		/* check if pipe reached left most pixel */
		if (pipe_x[pipe_start] <= -PIPE_WIDTH) {
			pipe_start++;
			score++;
			point_sound();
		}
		
		/* update bird */
		bird.y += speed;	
		/* update pipes */
		for (i = pipe_start; i < pipe_cnt; i++) {
			pipe_x[i] -= speed;
		}
		
		/* print everything to screen */
		lcd_clear();
		lcd_update();
		display_bird();
		for (i = pipe_start; i < pipe_cnt; i++) {
			display_pipe(i);
		}
		
		_delay_ms(75);
	}
}

/* shows the possible difficulties */
void difficulty_screen() {
	lcd_clear();
	lcd_update();
	
	lcd_goto_xy(10, 0);
	lcd_chr(ARROW_LEFT);
	lcd_str("Difficulty");
	lcd_chr(ARROW_RIGHT);
	
	lcd_goto_xy(0, 2);
	if (difficulty == EASY) {
		
		lcd_chr(ARROW_RIGHT);
		lcd_str("Easy");
	} else {
		lcd_str(" Easy");
	}
	
	lcd_goto_xy(0, 3);
	if (difficulty == MEDIUM) {
		lcd_chr(ARROW_RIGHT);
		lcd_str("Medium");
	} else {
		lcd_str(" Medium");
	}
	
	lcd_goto_xy(0, 4);
	if (difficulty == HARD) {
		lcd_chr(ARROW_RIGHT);
		lcd_str("Hard");
	} else {
		lcd_str(" Hard");
	}
	
	lcd_goto_xy(0,5);
	lcd_chr(ARROW_LEFT);
	lcd_str("Move");
	lcd_str("  ");
	lcd_str("Select");
	lcd_chr(ARROW_RIGHT);
}

/* wait for player to select difficulty */
void select_difficulty() {
	difficulty = 1;
	PIPE_SPACE = 25;
	PIPE_TURN = 55;
	difficulty_screen();
	int ok = 1;
	while(1) {
		/* check if player selected action */
		if (!(PIND & (1 << SELECT_BTN))) {
			if (difficulty == EASY) {
				PIPE_SPACE = 25;
				PIPE_TURN = 55;
			}
			if (difficulty == MEDIUM) {
				PIPE_SPACE = 20;
				PIPE_TURN = 45;
			}
			if (difficulty == HARD) {
				PIPE_SPACE = 15;
				PIPE_TURN = 35;
			}
			game();
			return;
		}
		
		/* check if player changes option */
		if (!(PIND & (1 << MOVE_BTN)) && ok) {
			ok = 0;
			if (difficulty == HARD)
				difficulty = EASY;
			else
				difficulty++;
			difficulty_screen();
			_delay_ms(250);
			ok = 1;
		}
		
	}
}

int main() {
	
	init_buttons();
	init_timer();
	
	lcd_init();
	lcd_clear();
	lcd_update();
	lcd_contrast(0x46);
	
	select_mode();
	
	return 0;
}